# -*- coding: utf-8 -*-
from __future__ import (
    division, absolute_import, print_function, unicode_literals,
)
from builtins import *                  # noqa
from future.builtins.disabled import *  # noqa

from magic_constraints.metadata import *  # noqa


def test_just_to_coverage():
    NAME
    VERSION
    AUTHORS
    EMAILS
    LICENSE
    URL
    DESCRIPTION
